#ifndef POLLY_SCOP_PROFILE_INFO_LOADER_H
#define POLLY_SCOP_PROFILE_INFO_LOADER_H
#include "polly/JITScopDetection.h"

#include "llvm/BasicBlock.h"
#include "llvm/InstrTypes.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Analysis/Passes.h"
#include "llvm/Analysis/ProfileInfo.h"
#include "llvm/Analysis/ProfileInfoLoader.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Support/CFG.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/Format.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/ADT/SmallSet.h"

#include "llvm/Analysis/RegionPass.h"
#include "llvm/Analysis/RegionInfo.h"
#include "llvm/Analysis/RegionIterator.h"

#include <set>
namespace polly {
  class ScopLoader : public FunctionPass, public ProfileInfo {
    static unsigned offset;

    std::string Filename;
    ProfileInfoLoader *PIL;
    JITScopDetection *JSD;
    RegionInfo *RI;

    // Has to be set once during analysis.
    std::pair<Function *, std::string> TopLevelKey;
  public:
    static char ID; // Class identification, replacement for typeinfo
    explicit ScopLoader(const std::string &filename = "")
      : FunctionPass(ID), Filename(filename) {

    }

    virtual void getAnalysisUsage(AnalysisUsage &AU) const {
      AU.addRequired<RegionInfo>();
      AU.addRequired<JITScopDetection>();
      AU.setPreservesAll();
    }

    virtual const char *getPassName() const {
      return "SCoP - Profiling information loader";
    }

    /// getAdjustedAnalysisPointer - This method is used when a pass implements
    /// an analysis interface through multiple inheritance.  If needed, it
    /// should override this to adjust the this pointer as needed for the
    /// specified pass info.
    virtual void *getAdjustedAnalysisPointer(AnalysisID PI) {
      if (PI == &ProfileInfo::ID)
        return (ProfileInfo*)this;
      return this;
    }

    virtual bool doInitialization(Module &M);
    virtual bool doFinalization(Module &M);

    /// run - Load the profile information from the specified file.
    virtual bool runOnFunction(Function &F);
  };
}

namespace llvm {
  class PassRegistry;
  void initializeScopLoaderPass(llvm::PassRegistry&);
}
#endif //POLLY_SCOP_PROFILE_LOADER_H
