#include <inttypes.h>
#include <papi.h>
#include <stdio.h>
#include <stdlib.h>

#define STACK_SIZE 1024

static long long stack[STACK_SIZE];
static long long time_stack[STACK_SIZE];
static long long *shadow_stack[STACK_SIZE];

static long long *total_time_arr = 0;

static int cur = -1;
static long long count = 0;

void papi_region_enter_scop(long long *arr, const char *dbg) {
  long long time = PAPI_get_virt_usec();

  if (!(cur < STACK_SIZE)) {
    fprintf(stderr, "ERROR(PAPI-Prof): Stack overflow!\n");
    exit(-2);
  }

  cur++;
  stack[cur]      = 0;
  time_stack[cur] = 0;

  (*arr)            -= time; 
  time_stack[cur]   -= time;
  shadow_stack[cur] = arr;
}

void papi_region_exit_scop(long long *arr, const char *dbg) {
  long long children_time = stack[cur];
  long long time = PAPI_get_virt_usec();
  long long elapsed = time + time_stack[cur];

  if (shadow_stack[cur] != arr) {
    fprintf(stderr, "ERROR(PAPI-Prof): popped wrong stack elem! (%s)\n", dbg);
    exit(-1);
  }

  if (children_time > elapsed) {
    fprintf(stderr, "ERROR(PAPI-Prof): Invalid time in stack! (%s)\n", dbg);
    exit(-3);
  } 
  
  (*arr) += time;
  (*arr) -= stack[cur];

  if (cur > 0)
    stack[cur-1] += time + time_stack[cur];

  cur--;
}

void papi_region_enter(long long *arr) {
  total_time_arr = arr;
  (*arr) -= PAPI_get_virt_usec();
}

void papi_region_exit(long long *arr) {
// This isn't always working for TLR of main,
// think of a reason why...
//  if (arr != total_time_arr)
//    (*arr) += PAPI_get_virt_usec();
}

void papi_atexit_handler(void) {
  /* Clean up the stack on close! */
  if (cur != -1) {
    int time = PAPI_get_virt_usec();
    while (cur != -1) {
      papi_region_exit_scop(shadow_stack[cur], "PAPI-Prof Cleanup.");
    }
  }

  (*total_time_arr) += PAPI_get_virt_usec();
}

void papi_region_setup(void) {
  int err = atexit(papi_atexit_handler);
  if (err)
    fprintf(stderr, "ERROR(PAPI-Prof): Failed to setup atexit handler (%d).\n", err);
}

