#!/usr/bin/env python
import argparse
import subprocess
import tempfile
import os
import sys
import traceback

PREFIX = '[PP] '
GCC = 'gcc'
CLANG = 'clang'
PLUGIN = '/home/simbuerg/opt/polly/lib/LLVMgold.so'

PROFILING_FLAGS = [
 '-lprofile_rt',
 '-lpapi'
]

PROFILING_PATH = [
 '-L/home/simbuerg/opt/polly/lib',
 '-L/home/simbuerg/opt/papi/lib'
]

LD_PATH = []
LD_FLAGS = []

def clang():
  global CLANG

  clng = CLANG
  # C or C++?
  if sys.argv[0].endswith('++') or '-cpp' in sys.argv:
    clng = 'clang++'

  return clng

def gcc():
  global GCC

  gcc = GCC

  # C or C++?
  if sys.argv[0].endswith('++') or '-cpp' in sys.argv:
    gcc = 'g++'

  return gcc

clean_log = True
# Execute process with nice pretty printing of our debugging outputs.
#
def log_exec(args, commandLine, msg, failOnErr = True, toFile = True):
  global clean_log
  level = len(traceback.format_stack())
  indent = '  ' * (level-5)

  if args.prg:
    print indent + '\033[1;31m[ '+ msg +' ]\033[1;m'
  
  if args.commands:
    print indent + '\033[1;36m' + ' '.join(commandLine) + '\033[1;m'

  if toFile:
    fn = getOutput(args) + '.pprof'
    if clean_log:
      f = open(fn, 'w')
      clean_log = False
    else:
      f = open(fn, 'a')

    f.write(' '.join(commandLine))
    f.write('\n')
    f.close()

  exit = subprocess.call(commandLine)
  if exit:
    print indent + '\033[1;36m' + msg + ' FAILED. This command did not work:\033[1;m'
    print indent + '\033[1;36m' + ' '.join(commandLine) + '\033[1;m'

    if failOnErr:
      sys.exit(1)

  return exit

# Preprocess the programm such that as many optimization opportunities
# as possible are exposed.
def optimize_ir(file, args):
  preoptFile = os.path.splitext(file)[0] + '.opt'

  # Sanity check our env
  if not 'LIBPOLLY' in os.environ:
    sys.exit('Polly library not provided. Please set LIBPOLLY environment ' + \
             'variable to the LLVMPolly.so file')

  polly = os.environ['LIBPOLLY']
 
  #preoptPasses = ['-basicaa'
  #              , '-mem2reg'
  #              , '-simplify-libcalls'
  #              , '-simplifycfg'
  #              , '-instcombine'
  #              , '-tailcallelim'
  #              , '-loop-simplify'
  #              , '-lcssa'
  #              , '-loop-rotate'
  #              , '-lcssa'
  #              , '-loop-unswitch'
  #              , '-instcombine'
  #              , '-loop-simplify'
  #              , '-lcssa'
  #              , '-indvars'
  #              , '-loop-deletion'
  #              , '-instcombine'
  #              , '-polly-prepare'
  #              , '-polly-region-simplify'
  #              , '-polly-independent'
  #              , '-indvars'
  # Not in pollycc , '-lowerinvoke'

  commandLine = [clang(), '-Xclang', '-load', '-Xclang', polly, '-O0', '-mllvm',
                          '-polly', '-S', '-emit-llvm'] + [file, '-o', preoptFile]
  #commandLine = ['opt', '-load', polly] + preoptPasses + [file, '-o', preoptFile]
  exit = log_exec(args, commandLine, 'Preoptimizing to expose optimization opportunities', False) 
  
  return preoptFile 

def link_ir(args, flags):
  outFile = getOutput(args)
  out = os.path.splitext(outFile)[0] + '.bc'
  global LD_PATH
  global PROFILING_FLAGS
  
  lflags = ['-l' + x for x in args.libraries] + ['-lprofile_rt'] + LD_FLAGS
  lpath = ['-L' + x for x in args.librarypath] + LD_PATH + ['-L/home/simbuerg/opt/polly/lib']

  # Order of linker flags MATTERS!!! don't swap args.files & args.unknown_args
  commandLine = [clang() , '-Wl,-plugin,'+PLUGIN+',-plugin-opt,emit-llvm',
                 '-emit-llvm', '-o', out] + args.files + lflags + lpath + \
                 args.unknown_args + LD_PATH + flags
  log_exec(args, commandLine, 'Linking BITCODE (unoptimized)')

  return out

def link_ir_fortran(args, flags):
  outFile = getOutput(args)
  out = os.path.splitext(outFile)[0] + '.bc'
  global LD_PATH
  global PROFILING_FLAGS
  
  lflags = ['-l' + x for x in args.libraries] + ['-lprofile_rt'] + LD_FLAGS
  lpath = ['-L' + x for x in args.librarypath] + LD_PATH + ['-L/home/simbuerg/opt/polly/lib']

  # Order of linker flags MATTERS!!! don't swap args.files & args.unknown_args
  #commandLine = ['gfortran' , '-fplugin=dragonegg', '-fplugin-arg-dragonegg-emit-ir',
  #               '-o', out] + args.files + lflags + lpath + \
  #               args.unknown_args + LD_PATH + flags
  # Order of linker flags MATTERS!!! don't swap args.files & args.unknown_args
  commandLine = [clang() , '-Wl,-plugin,'+PLUGIN+',-plugin-opt,emit-llvm',
                 '-emit-llvm', '-o', out] + args.files + lflags + lpath + \
                 args.unknown_args + LD_PATH + flags
  log_exec(args, commandLine, 'Linking human-readable bitcode (unoptimized)')

  commandLine = ['opt', out, '-o', out] 
  log_exec(args, commandLine, 'Creating bitcode')
  return out


def getOutput(args):
  if args.output and len(args.output) > 0:
    outFile = args.output
  else:
    outFile = 'a.out'

  return outFile

def link(ir, args, flags):
  global LD_PATH
  default_flags = ["-L/usr/lib/gcc/x86_64-linux-gnu/4.6",
                   "-L/usr/lib/gcc/x86_64-linux-gnu/4.6/../../../x86_64-linux-gnu",
                   "-L/lib/x86_64-linux-gnu",
                   "-L/lib/../lib64",
                   "-L/usr/lib/x86_64-linux-gnu/",
                   "-L/usr/lib/gcc/x86_64-linux-gnu/4.6/../../..",
                   "-L/lib",
                   "-L/usr/lib"]

  lflags = ['-l' + x for x in args.libraries] + PROFILING_FLAGS + LD_FLAGS
  lpath = ['-L' + x for x in args.librarypath] + LD_PATH + PROFILING_PATH

  out = getOutput(args)
  assembly = ir + '.s'
  commandLine = ['llc', '-O0', ir]
  log_exec(args, commandLine, 'Generating assembly')
 
  commandLine = [gcc(), assembly, '-o', out] + \
                 lflags + lpath + default_flags +\
                 LD_PATH

  log_exec(args, commandLine, 'Generating final output')
  return args.output

def link_fortran(ir, args, flags):
  global LD_PATH

  lflags = ['-l' + x for x in args.libraries] + PROFILING_FLAGS + LD_FLAGS
  lpath = ['-L' + x for x in args.librarypath] + LD_PATH + PROFILING_PATH

  out = getOutput(args)
  assembly = ir + '.s'
  commandLine = ['llc', ir]
  log_exec(args, commandLine, 'Generating assembly')
 
  commandLine = ['gfortran', assembly, '-o', out] + \
                 lflags + lpath + LD_PATH

  log_exec(args, commandLine, 'Generating final output')
  return args.output

def instrument_edge_profile(inFile, args):
  profFile = os.path.splitext(inFile)[0] + '.epbc'
  polly = os.environ['LIBPOLLY']
  commandLine = ['opt', '-load', polly,
                 '-insert-edge-profiling',
                 inFile, '-o', profFile]
  log_exec(args, commandLine, 'Inserting edge profiling')
  return profFile

def instrument_papi_profile(inFile, args):
  profFile = os.path.splitext(inFile)[0] + '.pfbc'
  polly = os.environ['LIBPOLLY']
  commandLine = ['opt', '-load', polly,
                 '-insert-papi-profiling',
                 inFile, '-o', profFile]
  log_exec(args, commandLine, 'Inserting papi profiling')
  return profFile
  
def instrument_papi_region_profile(inFile, args):
  profFile = os.path.splitext(inFile)[0] + '.prbc'
  polly = os.environ['LIBPOLLY']
  commandLine = ['opt', '-load', polly,
                 '-insert-papi-region-profiling',
                 inFile, '-o', profFile]
  log_exec(args, commandLine, 'Inserting papi region profiling')
  return profFile

# Instrument the input file.
#
def instrument_ir(inFile, args):
  ir = inFile
#  ir = instrument_papi_profile(ir, args)
#  ir = instrument_edge_profile(ir, args)
  ir = instrument_papi_region_profile(ir, args)

  return ir

